import os
import re
import subprocess
import sys
import tarfile
from fnmatch import fnmatch
from itertools import islice
from shutil import copyfile
from pathlib import Path

import click as click

_TARGET_PATTERN = re.compile(
    r'^(//)?{package}?:{target}$'.format(
        package=r'(?P<package>[^:/]+(/[^:/]+)*)',
        target=r'(?P<target>[^:/]+)'
    )
)
DEFAULT_QUERY = 'rdeps(..., kind(py_library, ...), 1) - attr(generator_function, py_proto_grpc_library, ...)'


class PkgTar:
    def __init__(self, target_name: str, package_name: str = ''):
        self.package_name = package_name
        self.target_name = target_name

    @staticmethod
    def parse(target: str):
        m = _TARGET_PATTERN.match(target)
        if not m:
            raise Exception('not a valid bazel target')

        return PkgTar(
            target_name=m.group('target'),
            package_name=m.group('package') or ''
        )

    @property
    def tar_path(self):
        if len(self.package_name):
            return f'bazel-bin/{self.package_name}/{self.target_name}.tar'
        else:
            return f'bazel-bin/{self.target_name}.tar'

    def __str__(self):
        return f'//{self.package_name}:{self.target_name}'


def _filter(path, name, include="*", exclude_path=None):
    return (
            not (include and not fnmatch(name, include)) and
            not (exclude_path and fnmatch(path, exclude_path))
    )


def _find(root, *args, **kwargs):
    for path, _, files in os.walk(root):
        for file_name in files:
            if _filter(path, file_name, *args, **kwargs):
                yield os.path.join(path, file_name)


def _find_parent_paths(folder, root="."):
    for n in Path(folder).parents:
        if n.samefile(root):
            break
        yield n


def _touch_init_file(folders, verbose=False):
    for folder in folders:
        init_file = Path(folder) / '__init__.py'
        if verbose:
            click.secho(f'+ {init_file}', fg='green')
        init_file.touch()
        yield init_file


def copy_files(verbose):
    click.secho(f'>>> Copy protobuf generated python files', fg='cyan')
    folders = set()
    for src in _find('bazel-bin', '*_pb2*.py', '*/external/*'):
        dst = src.replace('bazel-bin/', '').lstrip('/')
        if verbose:
            click.secho(f'/ {src} ==> {dst}', fg='blue')
        copyfile(src, dst)
        folders.update(_find_parent_paths(dst))

    for _ in _touch_init_file(folders, verbose):
        pass


def remove_old(verbose=False):
    click.secho(f'>>> Remove all previously generated files!', fg='red')
    for file_name in _find(root='.', include="*_pb2*.py"):
        if verbose:
            click.secho(f'- {file_name}', fg='red')
        os.remove(file_name)


def _run_command(commands, verbose=False, capture_output=False):
    if verbose:
        click.secho(f'$ {" ".join(commands)}')
    try:
        if capture_output:
            return subprocess.run(commands, capture_output=True, check=True)
        else:
            return subprocess.run(commands, stdout=subprocess.PIPE, check=True)
    except subprocess.CalledProcessError as e:
        click.secho(str(e))
        sys.exit(1)


def bazel_clean(verbose=False):
    click.secho(f'>>> Clean bazel', fg='cyan')
    _run_command(['bazel', 'clean', '--color', 'yes'], verbose)


def bazel_build(target, verbose=False):
    click.secho(f'>>> Build bazel target!', fg='cyan')
    if isinstance(target, str):
        target = [target]
    _run_command(['bazel', 'build', '--color', 'yes'] + target, verbose)


def bazel_query(query=DEFAULT_QUERY, verbose=False):
    click.secho(f'>>> Query bazel {query}', fg='cyan')
    result = _run_command(['bazel', 'query', query], verbose=verbose, capture_output=True)
    return result.stdout.decode("utf-8").strip().split("\n")


def list_py_in_tar(tar: tarfile.TarFile):
    for file_name in tar.getnames():
        if '..' not in file_name and fnmatch(file_name, "*_pb2*"):
            yield file_name


def add_to(tar_path: str, compression=None, verbose=False):
    click.secho(f'>>> Add protobuf generated python files to tar', fg='cyan')
    write_mode = f'w:{compression}' if compression else 'w'
    folders = set()
    with tarfile.open(tar_path, write_mode) as p_tar:
        for src in _find('bazel-bin', '*_pb2*.py', '*/external/*'):
            if verbose:
                click.secho(f'+ {tar_path} <== {src}', fg='green')
            p_tar.add(src, arcname=src[10:])
            folders.update(_find_parent_paths(src))
        for init_file in _touch_init_file(folders):
            if verbose:
                click.secho(f'+ {tar_path} <== {init_file}', fg='green')
            p_tar.add(init_file, arcname=str(init_file)[10:])


def extract_py(tar_path: str, verbose=False):
    click.secho(f'>>> Extract only protobuf generated python files from {tar_path}', fg='cyan')
    folders = set()
    with tarfile.open(tar_path, 'r') as tar:
        for file_name in list_py_in_tar(tar):
            if verbose:
                click.secho(f'x {tar_path} ==> {file_name}', fg='blue')
            tar.extract(file_name)
            folders.update(_find_parent_paths(file_name))
        for _ in _touch_init_file(folders, verbose):
            pass


def at_least_n(iterable, n):
    return len(list(islice(iterable, n))) > 0


def at_least_one(iterable):
    return at_least_n(iterable, 1)


@click.command(help=f'''
    This small script did the following:
      * Remove all existing *.pb2.py or *.pb2_grpc.py files
      * Bazel build the targets, by default it's coming from a query
      * Extract all *.pb2.py or *.pb2_grpc.py to current directoy.
      
    The query
      '{DEFAULT_QUERY}'
    will find all the reverse dependencies of `py_proto_grpc_library` genrules.

    
    In addition:
      * when --pack is defined, pack all the extracted files into a tar, which reduce the tar size significantly
      * when --extract is defined, skip the first two steps and extract directly from provided tar
    Note that --pack and --extract are mutually exclusive
    ''')
@click.option('-b', '--build', is_flag=True, help='rebuild the bazel target or not')
@click.option('-c', '--copy', is_flag=True, help='copy files out of bazel-bin')
@click.option('-e', '--extract-from', 'src_tar_path', default=None, type=str,
              help='the tarball to extract all the python from')
@click.option('--compression', default='gz', type=click.Choice(['', 'gz', 'bz2', 'xz']),
              help='set compression algorithm, if blank, no compression algorithm is applied')
@click.option('-o', '--output', 'tar_path', default="protos.tar", type=str,
              help='where to export the tarball')
@click.option('-p', '--pack', is_flag=True, help='if True will pack all the *_pb2*.py files into a tarball')
@click.option('-r', '--remove-only', is_flag=True, help='remove all previously generated python files and quit')
@click.option('-t', '--target', default='', type=str, help='the bazel target to generate the tarball')
@click.option('--query', default=DEFAULT_QUERY, type=str,
              help='the bazel query to find the relative targets, mutually exclusive to --target')
@click.option('-v', '--verbose', is_flag=True)
def cli(build, copy, compression, src_tar_path, pack, tar_path, remove_only, target, query, verbose):
    remove_old(verbose)
    if remove_only:
        exit(0)

    # if src_tar is designated explicitly, we extract from that tar only (remove first to ensure result)
    if src_tar_path:
        extract_py(src_tar_path, verbose)
        exit(0)

    if build or not at_least_one(_find('bazel-bin', '*_pb2*.py', '*/external/*')):
        if not target:
            target = bazel_query(query=query, verbose=verbose)
        bazel_build(target, verbose)

    if copy:
        copy_files(verbose)
        exit(0)

    if pack:
        add_to(tar_path, compression, verbose)
    else:
        copy_files(verbose)


if __name__ == '__main__':
    cli()
